﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using HeuristicLab.Grammars;

namespace CodeGenerator {
  public class BruteForceCodeGen {

    private string solverTemplate = @"
namespace ?PROBLEMNAME? {
  public sealed class ?IDENT?Solver {
    public class SearchTreeNode {
      public bool ready = false;
      public SearchTreeNode[] children;
      public int state;
      public int nextAltIdx;

      public SearchTreeNode(int state) {
        // do not initialize children yet to save mem
        nextAltIdx = 0;
      }
      public SearchTreeNode GetNextNode() {
        if(children == null) {
          int nChildren = Grammar.subtreeCount[state] == 1 ? Grammar.transitions[state].Length : Grammar.subtreeCount[state];
          children = new SearchTreeNode[nChildren];
        }
        if(children[nextAltIdx] == null) {
          children[nextAltIdx] = 
        }
      }
    }


    private readonly ?IDENT?Problem problem;
    private readonly Random random;
    private SearchTreeNode searchTree = null;

    private IEnumerable<Tree> GenerateTrees(int maxDepth) {
      if(searchTree==null) searchTree = new SearchTreeNode(0);
      
    }

    private Tree GenerateTree(SearchTreeNode node) {
      curDepth += 1;
      steps += 1;
      depth = Math.Max(depth, curDepth);
      Tree t = null;

      // terminals
      if(Grammar.subtreeCount[state] == 0) {
        t = CreateTerminalNode(state, random, problem);
      } else {
        // if the symbol has alternatives then we must choose one randomly (only one sub-tree in this case)
        if(Grammar.subtreeCount[state] == 1) {
          var targetStates = Grammar.transition[state];
          var altIdx = SampleAlternative(random, state, curDepth);
          var alternative = SampleTree(targetStates[altIdx], ref steps, ref curDepth, ref depth);
          t = new Tree(altIdx, new Tree[] { alternative });
        } else {
          // if the symbol contains only one sequence we must use create sub-trees for each symbol in the sequence
          Tree[] subtrees = new Tree[Grammar.subtreeCount[state]];
          for(int i = 0; i < Grammar.subtreeCount[state]; i++) {
            subtrees[i] = SampleTree(Grammar.transition[state][i], ref steps, ref curDepth, ref depth);
          }
          t = new Tree(-1, subtrees); // alternative index is ignored
        }
      }
      curDepth -=1;
      return t;
    }

    private static Tree CreateTerminalNode(int state, Random random, ?IDENT?Problem problem) {
      switch(state) {
        ?CREATETERMINALNODECODE?
        default: { throw new ArgumentException(""Unknown state index"" + state); }
      }
    }

    private int NextAlternative(SearchTreeNode node) {
?RETURNNEXTALTERNATIVECODE?
    }

    public static void Main(string[] args) {
      var problem = new ?IDENT?Problem();
      var solver = new ?IDENT?Solver(problem);
      solver.Start();
    }

    public ?IDENT?Solver(?IDENT?Problem problem) {
      this.problem = problem;
      this.random = new Random();
    }

    private void Start() {
      var bestF = ?MAXIMIZATION? ? double.NegativeInfinity : double.PositiveInfinity;
      int n = 0;
      long sumDepth = 0;
      long sumSize = 0;
      var sumF = 0.0;
      var sw = new System.Diagnostics.Stopwatch();
      sw.Start();
      for(int d = 1; d < 20; d++) {
        foreach(var t in GenerateTrees(d)) {
          var f = problem.Evaluate(t);
          
          n++;    
          sumF += f;
          if (problem.IsBetter(f, bestF)) {
            bestF = f;
            Console.WriteLine(""{0}\t{1}\t(size={2}, depth={3})"", n, bestF, 0, 0);
          }
          if (n % 1000 == 0) {
            sw.Stop();
            Console.WriteLine(""{0}\tbest: {1:0.000}\t(avg: {2:0.000})\t(avg size: {3:0.0})\t(avg. depth: {4:0.0})\t({5:0.00} sols/ms)"", n, bestF, sumF/1000.0, sumSize/1000.0, sumDepth/1000.0, 1000.0 / sw.ElapsedMilliseconds);
            sumSize = 0;
            sumDepth = 0;
            sumF = 0.0;
            sw.Restart();
          }
        }
      }
    }
  }
}";

    public void Generate(IGrammar grammar, IEnumerable<TerminalNode> terminals, bool maximization, SourceBuilder problemSourceCode) {
      var solverSourceCode = new SourceBuilder();
      solverSourceCode.Append(solverTemplate)
        .Replace("?MAXIMIZATION?", maximization.ToString().ToLowerInvariant())
        .Replace("?SAMPLEALTERNATIVECODE?", GenerateSampleAlternativeSource(grammar))
        .Replace("?CREATETERMINALNODECODE?", GenerateCreateTerminalCode(grammar, terminals))
      ;

      problemSourceCode.Append(solverSourceCode.ToString());
    }



    private string GenerateSampleAlternativeSource(IGrammar grammar) {
      Debug.Assert(grammar.Symbols.First().Equals(grammar.StartSymbol));
      var sb = new SourceBuilder();
      int stateCount = 0;
      foreach (var s in grammar.Symbols) {
        sb.AppendFormat("case {0}: ", stateCount++);
        if (grammar.IsTerminal(s)) {
          // ignore
        } else {
          var terminalAltIndexes = grammar.GetAlternatives(s)
            .Select((alt, idx) => new { alt, idx })
            .Where((p) => p.alt.All(symb => grammar.IsTerminal(symb)))
            .Select(p => p.idx);
          var nonTerminalAltIndexes = grammar.GetAlternatives(s)
            .Select((alt, idx) => new { alt, idx })
            .Where((p) => p.alt.Any(symb => grammar.IsNonTerminal(symb)))
            .Select(p => p.idx);
          var hasTerminalAlts = terminalAltIndexes.Any();
          var hasNonTerminalAlts = nonTerminalAltIndexes.Any();
          if (hasTerminalAlts && hasNonTerminalAlts) {
            sb.Append("if(random.NextDouble() < TerminalProbForDepth(depth)) {").BeginBlock();
            GenerateReturnStatement(terminalAltIndexes, sb);
            sb.Append("} else {");
            GenerateReturnStatement(nonTerminalAltIndexes, sb);
            sb.Append("}").EndBlock();
          } else {
            GenerateReturnStatement(grammar.NumberOfAlternatives(s), sb);
          }
        }
      }
      return sb.ToString();
    }

    private string GenerateCreateTerminalCode(IGrammar grammar, IEnumerable<TerminalNode> terminals) {
      Debug.Assert(grammar.Symbols.First().Equals(grammar.StartSymbol));
      var sb = new SourceBuilder();
      var allSymbols = grammar.Symbols.ToList();
      foreach (var s in grammar.Symbols) {
        if (grammar.IsTerminal(s)) {
          sb.AppendFormat("case {0}: {{", allSymbols.IndexOf(s)).BeginBlock();
          sb.AppendFormat("var t = new {0}Tree();", s.Name).AppendLine();
          var terminal = terminals.Single(t => t.Ident == s.Name);
          foreach (var constr in terminal.Constraints) {
            if (constr.Type == ConstraintNodeType.Set) {
              sb.Append("{").BeginBlock();
              sb.AppendLine("throw new NotImplementedException(\"Enumeration of terminal values is not implemented.\");");
              sb.AppendFormat("//var elements = problem.GetAllowed{0}_{1}().ToArray();", terminal.Ident, constr.Ident).AppendLine();
              sb.AppendFormat("//t.{0} = elements[random.Next(elements.Length)]; ", constr.Ident).EndBlock();
              sb.AppendLine("}");
            } else {
              sb.Append("{").BeginBlock();
              sb.AppendLine("throw new NotSupportedException(\"The brute force solver does not support RANGE constraints\");");
              sb.AppendLine("}");
            }
          }
          sb.AppendLine("return t;").EndBlock();
          sb.Append("}");
        }
      }
      return sb.ToString();
    }

    private void GenerateReturnStatement(IEnumerable<int> idxs, SourceBuilder sb) {
      if (idxs.Count() == 1) {
        sb.AppendFormat("return {0};", idxs.Single()).AppendLine();
      } else {
        var idxStr = idxs.Aggregate(string.Empty, (str, idx) => str + idx + ", ");
        sb.AppendFormat("return new int[] {{ {0} }}[random.Next({1})]; ", idxStr, idxs.Count()).AppendLine();
      }
    }

    private void GenerateReturnStatement(int nAlts, SourceBuilder sb) {
      if (nAlts > 1) {
        sb.AppendFormat("return random.Next({0});", nAlts).AppendLine();
      } else if (nAlts == 1) {
        sb.AppendLine("return 0; ");
      } else {
        sb.AppendLine("throw new InvalidProgramException();");
      }
    }
  }
}
